/* CmdBkg (c) 2016-2020 Mikael Sollenborn and Steffen Illhardt */
// Compilation: gcc -o cmdbkg.exe cmdbkg.c -lgdi32 -ldwmapi -lgdiplus -O3 -s

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#undef UNICODE
#define UNICODE
#include <sdkddkver.h>
#include <windows.h>
#include <dwmapi.h>

#define HEADER "CmdBkg v2.0 : Mikael Sollenborn and Steffen Illhardt 2016-2020"
#define SECOND_RUN L"_second?_"
#define DEFAULT_TRANSPARENCY 33
#define DELAY 20
#define TIMEOUT (DELAY * 100)
#define WM_CON_EVENT (WM_APP + 1)
#define STR_(v) #v
#define STR(m) STR_(m)

// Output the help message.
static inline void Help(FILE *stream)
{ /* clang-format off */
  fputs(
    "Usage: cmdbkg [\"file.[bmp|jpg|gif|png]\" [/t p] [/c bg] [/b] [/s]]\n"
    "  /t  Transparency with p = percentage 0..100 (default " STR(DEFAULT_TRANSPARENCY) ").\n"
    "  /c  Color of the background with bg = hex value format RRGGBB\n"
    "      (default is the color of the console buffer's upper left corner\n"
    "      at the time CmdBkg is called).\n"
    "  /b  Borders included.\n"
    "  /s  Stretch to fit.\n"
    "Specify no arguments to remove previous background.\n"
    "Specify /? as first argument to see this help message.\n\n",
    stream); /* clang-format on */
}

// Structure containing information about the passed program options.
typedef struct tag_Opt
{
  wchar_t *transparency; // pointer of the transparency definition in argList, will be NULL if /t is not passed
  wchar_t *color; // pointer of the color definition in argList, will be NULL if /c is not passed
  BOOL doBorders; // TRUE if /b is passed
  BOOL doStretch; // TRUE if /s is passed
} Opt;

// Structure containing data for the InsertBmpFromHandle threads.
typedef struct tag_ThreadArg
{
  const HBITMAP hBmp; // handle to the original bitmap
  const BITMAP *pBmp; // pointer to a BITMAP structure containing information about the original bitmap
  const BOOL stretch; // stretch to fit (TRUE) or center (FALSE)
  BOOL forceInsert; // copy of the global forceInsert value
  SIZE size; // width and height of the background to be used
  POINT previousPos; // previous position of the image on the background window
  SIZE previousSize; // previous size of the image on the background window
  HBITMAP hBmpScaled; // handle to the previously scaled image
} ThreadArg;

// Parse the passed program options.
BOOL GetOpt(int argc, wchar_t **argv, Opt *pOpt);
// Create the process for the background window.
BOOL CreateBkgProzess(const wchar_t *oldCmdLine, wchar_t **argList);
// Read the background color of a character cell and convert it from 4-bit RGBI (actually in the order IRGB) to 24-bit RGB.
DWORD GetConsoleBkgColor(void);
// Read the content of an image file, convert it to a bitmap in memory, and return a handle of it.
HBITMAP GetHBitmapFromFile(const wchar_t *imagePath, DWORD gesso, BITMAP *pBmp);
// Thred function to paint the image into the background window. It prevents the message queue from growing. Only one thread is running at a time.
DWORD WINAPI InsertBmpFromHandle(void *pArg);
// Find the window where the user data field contains the console window handle (previous background window, if any), and close it.
BOOL CALLBACK CloseWindowByUserData(HWND hwnd, LPARAM lPConWnd);
// Window callback function necessary to handle WM_DESTROY. (make the process leave the message loop)
LRESULT CALLBACK WindowProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam);
// Receive window events and post WM_CON_EVENT or WM_CLOSE to the message queue if the event came from the console window. (make the message loop iterate)
void CALLBACK WinEventProc(HWINEVENTHOOK hWinEventHook, DWORD event, HWND hwnd, LONG idObject, LONG idChild, DWORD idEventThread, DWORD dwmsEventTime);

HWND conWnd = NULL, bkgWnd = NULL; // console and background window handles need to be global because we can't pass them to WinEventProc
BOOL forceInsert = FALSE; // TRUE if WM_NCPAINT is processed in the message callback

int main(void)
{
  conWnd = GetConsoleWindow();
  if (!conWnd || !SendMessageW(conWnd, WM_GETICON, 0U, 0)) // the latter roughly verifies that the window is a console (rather than a terminal emulator)
  {
    MessageBoxA(NULL, "CmdBkg error: Unable to prepare window creation.", "Error", MB_ICONERROR);
    return 1;
  }

  int argCount = 0;
  const wchar_t *cmdLine = GetCommandLineW();
  wchar_t **argList = CommandLineToArgvW(cmdLine, &argCount);
  if (argList == NULL)
  {
    fputs("CmdBkg error: Unable to get arguments.\n", stderr);
    return 1;
  }

  if (argCount > 1 && wcscmp(argList[1], L"/?") == 0) // show help
  {
    puts(HEADER);
    Help(stdout);
    LocalFree(argList);
    return 0;
  }

  SetLayeredWindowAttributes(conWnd, 0UL, 0xFF, LWA_ALPHA); // opaque (if the window isn't layered this call will fail without downside because it's already opaque)
  EnumWindows(CloseWindowByUserData, (LPARAM)conWnd); // close the background window running for the monitored console (if any)
  if (argCount <= 1) // no argument passed => quit without background window
  {
    LocalFree(argList);
    return 0;
  }

  if (wcscmp(argList[1], SECOND_RUN) != 0) // create a new process for the background window, and quit
  {
    SetWindowLongPtrW(conWnd, GWL_EXSTYLE, GetWindowLongPtrW(conWnd, GWL_EXSTYLE) | WS_EX_LAYERED); // transparency can only be updated if the window is layered
    if (!CreateBkgProzess(cmdLine, argList))
    {
      fputs("CmdBkg error: Creation of background process failed.\n", stderr);
      LocalFree(argList);
      return 1;
    }

    LocalFree(argList);
    return 0;
  }

  // everything from this point onwards is for the background process only
  Opt opt = { 0 };
  if (!GetOpt(argCount, argList, &opt))
  {
    fputs("CmdBkg syntax error.\n", stderr);
    Help(stderr);
    LocalFree(argList);
    return 1;
  }

  const DWORD gesso = opt.color ? wcstoul(opt.color, NULL, 16) & 0xFFFFFF : GetConsoleBkgColor(); // order 00RRGGBB (ARGB)
  WNDCLASSW wndCls = { 0 };
  wndCls.lpszClassName = L"CmdBkgWindowBackground";
  wndCls.lpfnWndProc = WindowProc;
  wndCls.hbrBackground = CreateSolidBrush(((gesso & 0xFF0000) >> 16) | (gesso & 0xFF00) | ((gesso & 0xFF) << 16)); // order 00BBGGRR (COLORREF)
  wndCls.hCursor = LoadCursorW(NULL, IDC_ARROW);
  if (!RegisterClassW(&wndCls) || !(bkgWnd = CreateWindowExW(WS_EX_TOOLWINDOW, wndCls.lpszClassName, NULL, WS_POPUP | WS_VISIBLE, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, NULL, NULL, NULL, NULL)))
  {
    fputs("CmdBkg error: Unable to create background window.", stderr);
    LocalFree(argList);
    return 1;
  }

  SetWindowLongPtrW(bkgWnd, GWLP_USERDATA, (LONG_PTR)conWnd); // the user data field contains the console window handle for the sake of finding it from within another instance of cmdbkg
  FreeConsole(); // don't move this line to a position before GetConsoleBkgColor is still called, or before the user data field contains the console window handle because WaitForInputIdle will return after this line
  BITMAP bmp = { 0 };
  const HBITMAP hBmp = GetHBitmapFromFile(argList[2], gesso, &bmp);
  const unsigned long percentage = opt.transparency ? wcstoul(opt.transparency, NULL, 10) : DEFAULT_TRANSPARENCY;
  LocalFree(argList); // don't move this line to a position before opt.transparency or opt.color is still used, they point to elements in argList
  if (!hBmp)
  {
    MessageBoxA(NULL, "CmdBkg error: Unable to load image.", "Error", MB_ICONERROR);
    return 1;
  }

  const HWINEVENTHOOK hHook = SetWinEventHook(EVENT_OBJECT_DESTROY, EVENT_OBJECT_LOCATIONCHANGE, NULL, WinEventProc, 0UL, 0UL, WINEVENT_OUTOFCONTEXT | WINEVENT_SKIPOWNPROCESS);
  if (!hHook)
  {
    MessageBoxA(NULL, "CmdBkg error: Unable to hook up the console window.", "Error", MB_ICONERROR);
    DeleteObject(hBmp);
    return 1;
  }

  RECT wndBounds = { 0 }, extendBounds = { 0 };
  long frameWidth = 0L, frameHight = 0L, frameTitleHight = 0L;
  GetWindowRect(conWnd, &wndBounds);
  DwmGetWindowAttribute(conWnd, DWMWA_EXTENDED_FRAME_BOUNDS, &extendBounds, sizeof(extendBounds));
  if (!opt.doBorders || (opt.doBorders && extendBounds.left != wndBounds.left))
  {
    frameWidth = GetSystemMetrics(SM_CXSIZEFRAME);
    if (!opt.doBorders)
      frameTitleHight = GetSystemMetrics(SM_CYSIZEFRAME) + GetSystemMetrics(SM_CYCAPTION);

    frameHight = frameTitleHight + GetSystemMetrics(SM_CYSIZEFRAME);
  }

  SetForegroundWindow(conWnd);
  SetLayeredWindowAttributes(conWnd, 0UL, (BYTE)((255UL * (100UL - (percentage > 100UL ? DEFAULT_TRANSPARENCY : percentage)) + 50UL) / 100UL), LWA_ALPHA); // transparent
  BOOL redraw = TRUE, bkgIsIconic = FALSE;
  UINT_PTR timerId = 0U;
  ThreadArg tArg = { hBmp, &bmp, opt.doStretch, FALSE, { 0 }, { -1L, -1L }, { -1L, -1L }, NULL };
  HANDLE hThread = NULL;
  MSG msg = { 0 };
  while (GetMessageW(&msg, NULL, 0U, 0U))
  {
    if (msg.message == WM_CON_EVENT)
      redraw = TRUE;
    else if (msg.message == WM_TIMER && msg.wParam == timerId) // ready to restore the background window
    {
      KillTimer(NULL, timerId);
      timerId = 0U;
      bkgIsIconic = FALSE; // not yet, but later within this iteration
    }
    else
      DispatchMessageW(&msg);

    if (!bkgIsIconic)
    {
      if (redraw)
      {
        GetWindowRect(conWnd, &wndBounds);
        SIZE wndSize = { wndBounds.right - wndBounds.left - frameWidth * 2, wndBounds.bottom - wndBounds.top };
        SetWindowPos(bkgWnd, conWnd, wndBounds.left + frameWidth, wndBounds.top + frameTitleHight, wndSize.cx, wndSize.cy - frameHight, SWP_SHOWWINDOW | SWP_NOACTIVATE);
        if (IsIconic(conWnd))
          bkgIsIconic = TRUE; // SetWindowPos did it
        else if (!hThread || WaitForSingleObject(hThread, 0UL) != WAIT_TIMEOUT) // to keep the window responsive skip drawing the image if we are still about to draw
        {
          if (hThread)
            CloseHandle(hThread);

          tArg.forceInsert = forceInsert; // we need a copy to avoid data races
          tArg.size = wndSize; // same here
          tArg.size.cy -= frameTitleHight;
          hThread = CreateThread(NULL, 0u, InsertBmpFromHandle, &tArg, 0UL, NULL);
          redraw = FALSE;
          forceInsert = FALSE;
        }
      }
    }
    else if (!timerId && !IsIconic(conWnd))
      timerId = SetTimer(NULL, 0U, DELAY * 10, NULL); // let the console fade in before we restore the background, but don't block the message loop while we are waiting
  }

  UnhookWinEvent(hHook);
  if (hThread)
    CloseHandle(hThread);

  if (tArg.hBmpScaled)
    DeleteObject(tArg.hBmpScaled);

  DeleteObject(hBmp);
  return 0;
}

// GDI+ headers are C++-only in MS Visual Studio (even dependencies of the Flat API). Instead of including <gdiplus.h> we declare the necessary things here. However the gdiplus lib has to be linked.
typedef struct tag_GdiplusStartupInput
{
  UINT32 GdiplusVersion;
  void *DebugEventCallback;
  BOOL SuppressBackgroundThread;
  BOOL SuppressExternalCodecs;
} GdiplusStartupInput;

int __stdcall GdiplusStartup(ULONG_PTR *, const GdiplusStartupInput *, void *);
void __stdcall GdiplusShutdown(ULONG_PTR);
int __stdcall GdipCreateBitmapFromFile(const wchar_t *, void **);
int __stdcall GdipCreateHBITMAPFromBitmap(void *, HBITMAP *, DWORD);

BOOL GetOpt(int argc, wchar_t **argv, Opt *pOpt) // relies on the referenced Opt object being zero-initialized
{
  for (int i = 3; i < argc; ++i) // argv[2] is the image file's path at the point where we call this function
  {
    if (*argv[i] == L'/')
    {
      if (_wcsicmp(L"s", &argv[i][1]) == 0)
        pOpt->doStretch = TRUE;
      else if (_wcsicmp(L"b", &argv[i][1]) == 0)
        pOpt->doBorders = TRUE;
      else if (_wcsicmp(L"t", &argv[i][1]) == 0)
      {
        if (i + 1 == argc || *argv[i + 1] == L'/')
          return FALSE;

        pOpt->transparency = argv[++i];
      }
      else if (_wcsicmp(L"c", &argv[i][1]) == 0)
      {
        if (i + 1 == argc || *argv[i + 1] == L'/')
          return FALSE;

        pOpt->color = argv[++i];
      }
      else
        return FALSE;
    }
  }

  return (argc == (3 + pOpt->doStretch + pOpt->doBorders + ((pOpt->color != NULL) << 1) + ((pOpt->transparency != NULL) << 1)));
}

BOOL CreateBkgProzess(const wchar_t *oldCmdLine, wchar_t **argList)
{
  const size_t oldLen = wcslen(oldCmdLine);
  wchar_t *newCmdLine = malloc(oldLen * sizeof(wchar_t) + sizeof(SECOND_RUN) + sizeof(wchar_t));
  if (!newCmdLine)
    return FALSE;

  const wchar_t *oldCursor = (*oldCmdLine == L'"' ? oldCmdLine + 2 : oldCmdLine) + wcslen(argList[0]);
  wchar_t *newCursor = (wchar_t *)memcpy(newCmdLine, oldCmdLine, (size_t)(oldCursor - oldCmdLine) * sizeof(wchar_t)) + (oldCursor - oldCmdLine);
  *newCursor++ = L' ';
  newCursor = (wchar_t *)memcpy(newCursor, SECOND_RUN, sizeof(SECOND_RUN) - sizeof(wchar_t)) + (sizeof(SECOND_RUN) - sizeof(wchar_t)) / sizeof(wchar_t);
  memcpy(newCursor, oldCursor, (oldLen - (oldCursor - oldCmdLine) + 1) * sizeof(wchar_t));

  PROCESS_INFORMATION pi = { 0 };
  STARTUPINFOW si = { .cb = sizeof(STARTUPINFOW) };
  if (!CreateProcessW(NULL, newCmdLine, NULL, NULL, FALSE, 0UL, NULL, NULL, &si, &pi)) // start the child process
  {
    free(newCmdLine);
    return FALSE;
  }

  WaitForInputIdle(pi.hProcess, TIMEOUT); // ensure the child process' window can already be found before we return to our parent
  CloseHandle(pi.hProcess);
  CloseHandle(pi.hThread);
  free(newCmdLine);
  return TRUE;
}

DWORD GetConsoleBkgColor(void)
{
  const HANDLE hOut = CreateFileW(L"CONOUT$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0UL, NULL); // because stdout could be redirected ...
  if (hOut != INVALID_HANDLE_VALUE)
  {
    WORD attr = 0;
    const BOOL good = ReadConsoleOutputAttribute(hOut, &attr, 1UL, (COORD){ 0 }, (DWORD[]){ 0 });
    CloseHandle(hOut);
    if (good)
    { // 4-Bit RGBI, converted using 0xAA (2/3 0xFF) for dark colors and 0x55 (1/3 0xFF) added for brightness to make "dark white" lighter than "bright black"
      const DWORD bright = (attr & BACKGROUND_INTENSITY) ? 0x55UL : 0x00UL;
      return ((((attr & BACKGROUND_RED) ? 0xAA : 0x00) + bright) << 16) | ((((attr & BACKGROUND_GREEN) ? 0xAA : 0x00) + bright) << 8) | (((attr & BACKGROUND_BLUE) ? 0xAA : 0x00) + bright);
    }
  }

  return 0UL;
}

HBITMAP GetHBitmapFromFile(const wchar_t *imagePath, DWORD gesso, BITMAP *pBmp)
{
  HBITMAP hBmp = NULL;
  const GdiplusStartupInput gpStartupInput = { 1U, NULL, FALSE, FALSE };
  ULONG_PTR gpToken = 0U;
  GdiplusStartup(&gpToken, &gpStartupInput, NULL);
  void *bmpPtr = NULL;
  GdipCreateBitmapFromFile(imagePath, &bmpPtr);
  if (bmpPtr)
    GdipCreateHBITMAPFromBitmap(bmpPtr, &hBmp, gesso);

  GdiplusShutdown(gpToken);
  if (hBmp && !GetObjectW(hBmp, sizeof(BITMAP), pBmp))
  {
    DeleteObject(hBmp);
    return NULL;
  }

  if (hBmp && pBmp->bmWidth > GetSystemMetrics(SM_CXFULLSCREEN) && pBmp->bmHeight > GetSystemMetrics(SM_CYFULLSCREEN))
  { // if the original image is larger than the screen, shrink it to make scaling in InsertBmpFromHandle cheaper
    const double factorW = (double)GetSystemMetrics(SM_CXFULLSCREEN) / pBmp->bmWidth, factorH = (double)GetSystemMetrics(SM_CYFULLSCREEN) / pBmp->bmHeight, factor = factorW > factorH ? factorW : factorH;
    hBmp = (HBITMAP)CopyImage((HANDLE)hBmp, IMAGE_BITMAP, (int)(pBmp->bmWidth * factor + .5), (int)(pBmp->bmHeight * factor + .5), LR_COPYDELETEORG);
    if (hBmp && !GetObjectW(hBmp, sizeof(BITMAP), pBmp))
    {
      DeleteObject(hBmp);
      return NULL;
    }
  }

  return hBmp;
}

DWORD WINAPI InsertBmpFromHandle(void *pArg)
{
  ThreadArg *pTArg = pArg;
  POINT pos = { 0 };
  const SIZE totalSize = pTArg->size;
  int ratioType = 0;
  if (!pTArg->stretch)
  {
    ratioType = (double)pTArg->size.cy / pTArg->size.cx < (double)pTArg->pBmp->bmHeight / pTArg->pBmp->bmWidth ? 1 : 2;
    if (ratioType == 1)
    {
      pTArg->size.cx = (long)(totalSize.cx * ((double)pTArg->pBmp->bmWidth / pTArg->pBmp->bmHeight) * ((double)pTArg->size.cy / totalSize.cx) + .5);
      pos.x = (long)(((double)totalSize.cx - pTArg->size.cx) / 2 + .5);
    }
    else
    {
      pTArg->size.cy = (long)(totalSize.cy * ((double)pTArg->pBmp->bmHeight / pTArg->pBmp->bmWidth) * ((double)pTArg->size.cx / totalSize.cy) + .5);
      pos.y = (long)(((double)totalSize.cy - pTArg->size.cy) / 2 + .5);
    }
  }

  if (!pTArg->hBmpScaled || (!pTArg->stretch && memcmp(&pos, &pTArg->previousPos, sizeof(POINT))) || (pTArg->stretch && memcmp(&pTArg->size, &pTArg->previousSize, sizeof(SIZE))))
  {
    if (pTArg->hBmpScaled)
      DeleteObject(pTArg->hBmpScaled);

    pTArg->hBmpScaled = (HBITMAP)CopyImage((HANDLE)pTArg->hBmp, IMAGE_BITMAP, (int)pTArg->size.cx, (int)pTArg->size.cy, LR_DEFAULTCOLOR);
    if (!pTArg->hBmpScaled)
      return 0UL;

    if (ratioType == 1)
    {
      RedrawWindow(bkgWnd, &(RECT){ 0L, 0L, pos.x, pTArg->size.cy }, NULL, RDW_ERASE | RDW_INVALIDATE); // redraw only uncovered areas to avoid flicker
      RedrawWindow(bkgWnd, &(RECT){ pos.x + pTArg->size.cx, 0L, totalSize.cx, pTArg->size.cy }, NULL, RDW_ERASE | RDW_INVALIDATE | RDW_UPDATENOW);
    }
    else if (ratioType == 2)
    {
      RedrawWindow(bkgWnd, &(RECT){ 0L, 0L, pTArg->size.cx, pos.y }, NULL, RDW_ERASE | RDW_INVALIDATE);
      RedrawWindow(bkgWnd, &(RECT){ 0L, pos.y + pTArg->size.cy, pTArg->size.cx, totalSize.cy }, NULL, RDW_ERASE | RDW_INVALIDATE | RDW_UPDATENOW);
    }

    pTArg->forceInsert = TRUE; // we always want to draw the bitmap if we had to rescale it
    pTArg->previousPos = pos;
    pTArg->previousSize = pTArg->size;
  }

  if (pTArg->forceInsert)
  {
    const HDC hDc = GetDC(bkgWnd);
    if (hDc)
    {
      const HDC hDcBmp = CreateCompatibleDC(hDc);
      if (hDcBmp)
      {
        const HGDIOBJ hGdiObj = SelectObject(hDcBmp, (HGDIOBJ)pTArg->hBmpScaled);
        if (hGdiObj != NULL)
        {
          BitBlt(hDc, (int)pos.x, (int)pos.y, (int)pTArg->size.cx, (int)pTArg->size.cy, hDcBmp, 0, 0, SRCCOPY);
          pTArg->hBmpScaled = (HBITMAP)SelectObject(hDcBmp, hGdiObj);
          if (pTArg->hBmpScaled == HGDI_ERROR)
            pTArg->hBmpScaled = NULL;
        }
        else
        {
          DeleteObject(pTArg->hBmpScaled);
          pTArg->hBmpScaled = NULL;
        }

        DeleteDC(hDcBmp);
      }

      ReleaseDC(bkgWnd, hDc);
    }
  }

  return 0UL;
}

BOOL CALLBACK CloseWindowByUserData(HWND hwnd, LPARAM lPConWnd)
{
  if (GetWindowLongPtrW(hwnd, GWLP_USERDATA) != lPConWnd)
    return TRUE;

  SendNotifyMessageW(hwnd, WM_CLOSE, 0U, 0);
  for (unsigned i = 0U; i++ < TIMEOUT / DELAY && (IsWindow(hwnd) || IsIconic(hwnd)); Sleep(DELAY)) // ensure the window has been closed before we return
    ;

  return FALSE;
}

LRESULT CALLBACK WindowProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
  if (msg == WM_DESTROY)
  {
    PostQuitMessage(0);
    return 0;
  }

  if (msg == WM_NCPAINT) // if the window was partially moved across the edges of the screen, the parts of the picture which have been invisible won't get restored
    forceInsert = TRUE; // indicate to force painting the picture new in this case

  return DefWindowProcW(hwnd, msg, wparam, lparam);
}

void CALLBACK WinEventProc(HWINEVENTHOOK hHook, DWORD event, HWND hwnd, LONG obj, LONG child, DWORD thread, DWORD time)
{
  hHook = hHook, obj = obj, child = child, thread = thread, time = time; // unused
  if (conWnd != hwnd)
    return;

  if (event == EVENT_OBJECT_LOCATIONCHANGE || event == EVENT_OBJECT_FOCUS)
    PostMessageW(bkgWnd, WM_CON_EVENT, 0U, 0);
  else if (event == EVENT_OBJECT_DESTROY)
    PostMessageW(bkgWnd, WM_CLOSE, 0U, 0);
}
